/////////////////////////////////////////////////////////////
// Name   : hidconfig.c
// Desc   : configuration tool for Trigger module
// Author : Atsushi Sato (NEKORISU)
// Date   : 2017-10-26
/////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <usb.h>

////////////////////////////////////////////
// Definition supported key table
////////////////////////////////////////////
typedef	struct	{
	char			*key_name;
	unsigned char	key_code;
}KEY_TABLE;
KEY_TABLE tblKey[] = {
	{"KEY_A",		4},
	{"KEY_B",		5},
	{"KEY_C",		6},
	{"KEY_D",		7},
	{"KEY_E",		8},
	{"KEY_F",		9},
	{"KEY_G",		10},
	{"KEY_H",		11},
	{"KEY_I",		12},
	{"KEY_J",		13},
	{"KEY_K",		14},
	{"KEY_L",		15},
	{"KEY_M",		16},
	{"KEY_N",		17},
	{"KEY_O",		18},
	{"KEY_P",		19},
	{"KEY_Q",		20},
	{"KEY_R",		21},
	{"KEY_S",		22},
	{"KEY_T",		23},
	{"KEY_U",		24},
	{"KEY_V",		25},
	{"KEY_W",		26},
	{"KEY_X",		27},
	{"KEY_Y",		28},
	{"KEY_Z",		29},
	{"KEY_1",		30},
	{"KEY_2",		31},
	{"KEY_3",		32},
	{"KEY_4",		33},
	{"KEY_5",		34},
	{"KEY_6",		35},
	{"KEY_7",		36},
	{"KEY_8",		37},
	{"KEY_9",		38},
	{"KEY_0",		39},
	{"KEY_F1",		58},
	{"KEY_F2",		59},
	{"KEY_F3",		60},
	{"KEY_F4",		61},
	{"KEY_F5",		62},
	{"KEY_F6",		63},
	{"KEY_F7",		64},
	{"KEY_F8",		65},
	{"KEY_F9",		66},
	{"KEY_F10",		67},
	{"KEY_F11",		68},
	{"KEY_F12",		69}
};

KEY_TABLE	tblModifiedKey[] = {
	{"CTRL_LEFT",		(1<<0)},
	{"SHIFT_LEFT",		(1<<1)},
	{"ALT_LEFT",		(1<<2)},
	{"CTRL_RIGHT",		(1<<4)},
	{"SHIFT_RIGHT",		(1<<5)},
	{"ALT_RIGHT",		(1<<6)}
};

#define	KEY_MAP_MAX		4
typedef	struct	{
	unsigned char	keycode;
	unsigned char	modifiedKey;
}KEY_MAP;
KEY_MAP		tblKeyMap[KEY_MAP_MAX];

#define	ishexdigit(c)	(( ((c) >= '0' && (c) <= '9') || (toupper(c) >= 'A' && toupper(c) <= 'F') ) ? 1 : 0)
#define	hextodec(c)		( ((c) >= '0' && (c) <= '9') ? ((c) - '0') : (10 + (c) - 'A'))

#define USBRQ_HID_GET_REPORT    0x01
#define USBRQ_HID_SET_REPORT    0x09

#define USB_HID_REPORT_TYPE_FEATURE 3

#define	VENDOR_ID		0x20A0
#define	PRODUCT_ID		0x4282
#define	VENDOR_NAME		"NEKORISU.com"
#define	DEVICE_NAME		"USB-Trigger"


#define USBOPEN_SUCCESS         0   /* no error */
#define USBOPEN_ERR_ACCESS      1   /* not enough permissions to open device */
#define USBOPEN_ERR_IO          2   /* I/O error */
#define USBOPEN_ERR_NOTFOUND    3   /* device not found */

typedef struct usbDevice    usbDevice_t;

typedef	struct	{
	unsigned char	reportId;
	unsigned char	key[4][2];
} __attribute__((packed)) CONFIG_DATA;

//////////////////////////////////////////////////////////////
// Get strings from DESCRIPTOR
//////////////////////////////////////////////////////////////
// usbhidGetStringAscii
//////////////////////////////////////////////////////////////
static int usbhidGetStringAscii(usb_dev_handle *dev, int index, char *buf, int buflen)
{
	char    buffer[256];
	int     rval, i;

	if((rval = usb_get_string_simple(dev, index, buf, buflen)) >= 0)
		return rval;

	if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, 0x0409, buffer, sizeof(buffer), 5000)) < 0)
		return rval;
	if(buffer[1] != USB_DT_STRING) {
		*buf = 0;
		return 0;
	}
	if((unsigned char)buffer[0] < rval)
		rval = (unsigned char)buffer[0];
	rval /= 2;

    /* lossy conversion to ISO Latin1: */
	for(i = 1; i < rval; i++) {
		if(i > buflen)
			break;
		buf[i-1] = buffer[2 * i];
		if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
			buf[i-1] = '?';
	}
	buf[i-1] = 0;

	return i-1;
}

//////////////////////////////////////////////////////////////
// Open HID device
//////////////////////////////////////////////////////////////
// usbhidOpenDevice
//////////////////////////////////////////////////////////////
int usbhidOpenDevice(usbDevice_t **device, int vendor, char *vendorName, int product, char *productName)
{
	struct usb_bus      *bus;
	struct usb_device   *dev;
	usb_dev_handle      *handle = NULL;
	int                 errorCode = USBOPEN_ERR_NOTFOUND;
	static int          didUsbInit = 0;

	if(!didUsbInit) {
		usb_init();
		didUsbInit = 1;
    }

    usb_find_busses();
	usb_find_devices();
	for(bus=usb_get_busses(); bus; bus=bus->next) {
		for(dev=bus->devices; dev; dev=dev->next) {
			if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product) {
				char    string[256];
				int     len;

				handle = usb_open(dev); /* we need to open the device in order to query strings */
				if(!handle){
					errorCode = USBOPEN_ERR_ACCESS;
					fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
					continue;
				}
				if(vendorName == NULL && productName == NULL) {  /* name does not matter */
					break;
				}

				/* now check whether the names match: */
				len = usbhidGetStringAscii(handle, dev->descriptor.iManufacturer, string, sizeof(string));
				if(len < 0) {
					errorCode = USBOPEN_ERR_IO;
					fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
				}else{
					errorCode = USBOPEN_ERR_NOTFOUND;
					if(strcmp(string, vendorName) == 0) {
						len = usbhidGetStringAscii(handle, dev->descriptor.iProduct, string, sizeof(string));
						if(len < 0) {
							errorCode = USBOPEN_ERR_IO;
							fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
						}else{
							errorCode = USBOPEN_SUCCESS;
							if(strcmp(string, productName) == 0) {
								if( usb_set_configuration(handle, dev->config->bConfigurationValue) < 0 ) {
									if( usb_detach_kernel_driver_np(handle, dev->config->interface->altsetting->bInterfaceNumber) < 0 ) {
										fprintf(stderr, "usb_set_configuration Error.\n");
										fprintf(stderr, "usb_detach_kernel_driver_np Error.(%s)\n", usb_strerror());
									}
								}
								if( usb_claim_interface(handle, dev->config->interface->altsetting->bInterfaceNumber) < 0 ) {
									if( usb_detach_kernel_driver_np(handle, dev->config->interface->altsetting->bInterfaceNumber) < 0 ) {
										fprintf(stderr, "usb_claim_interface Error.\n");
										fprintf(stderr, "usb_detach_kernel_driver_np Error.(%s)\n", usb_strerror());
									}
								}

								if( usb_claim_interface(handle, dev->config->interface->altsetting->bInterfaceNumber) < 0 ) {
									fprintf(stderr, "usb_claim_interface Error.(%s)\n", usb_strerror());
								}
								break;
							}
						}
					}
				}
				usb_close(handle);
				handle = NULL;
			}
		}
		if(handle)
			break;
	}
	if(handle != NULL) {
		errorCode = 0;
		*device = (void *)handle;
	}

	return errorCode;
}

//////////////////////////////////////////////////////////////
// Close HID device
//////////////////////////////////////////////////////////////
// usbhidCloseDevice
//////////////////////////////////////////////////////////////
void    usbhidCloseDevice(usbDevice_t *device)
{
	if(device != NULL)
		usb_close((void *)device);
}

//////////////////////////////////////////////////////////////
// Set Report
//////////////////////////////////////////////////////////////
// usbhidSetReport
//////////////////////////////////////////////////////////////
int usbhidSetReport(usbDevice_t *device, char *buffer, int len)
{
	int bytesSent;
	int	reportId = buffer[0];

	bytesSent = usb_control_msg((void *)device, USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_OUT, USBRQ_HID_SET_REPORT, USB_HID_REPORT_TYPE_FEATURE << 8 | (reportId & 0xff), 0, buffer, len, 5000);

	return bytesSent;
}

//////////////////////////////////////////////////////////////
// Get Report
//////////////////////////////////////////////////////////////
// usbhidGetReport
//////////////////////////////////////////////////////////////
int usbhidGetReport(usbDevice_t *device, char *buffer, int len)
{
	int bytesReceived;
	int	reportId = buffer[0];

	bytesReceived = usb_control_msg((void *)device, USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_IN, USBRQ_HID_GET_REPORT, USB_HID_REPORT_TYPE_FEATURE << 8 | (reportId & 0xff), 0, buffer, len, 5000);

	return bytesReceived;
}

//////////////////////////////////////////////////////////////
// Search specified code from key tables
//////////////////////////////////////////////////////////////
// matchKeyCode
//////////////////////////////////////////////////////////////
char *matchKeyCode(char code, KEY_TABLE *tbl, int numTable)
{
	int				i;
	char			*keyname = NULL;

	for(i = 0; i < numTable; i++) {
		if( tbl[i].key_code == code ) {
			keyname = tbl[i].key_name;
			break;
		}
	}

	return keyname;
}

//////////////////////////////////////////////////////////////
// Search specified key name from key tables
//////////////////////////////////////////////////////////////
// matchKeyName
//////////////////////////////////////////////////////////////
unsigned char matchKeyName(char *keyname, KEY_TABLE *tbl, int numTable)
{
	unsigned char	retc = 0xff;
	int				i;

	for(i = 0; i < numTable; i++) {
		if(strlen(keyname) == strlen(tbl[i].key_name)) {
			if(strcasecmp(keyname, tbl[i].key_name) == 0 ) {
				retc = tbl[i].key_code;
				break;
			}
		}
	}

	return retc;
}

//////////////////////////////////////////////////////////////
// Get specified code from table or directly
//////////////////////////////////////////////////////////////
unsigned char	getSpecifiedCode(char *str, KEY_TABLE *tbl, int numTable)
{
	unsigned char	code = 0xFF;

	if( memcmp(str, "0x", 2) == 0 ) {
		if( (strlen(str) == 4) && ishexdigit(str[2]) && ishexdigit(str[3]) ) {
			code = hextodec(str[2]) * 16 + hextodec(str[3]);
		}
	}else{
		code = matchKeyName(str, tbl, numTable);
	}

	return code;
}

//////////////////////////////////////////////////////////////
// Get specified Key name from table or directly
//////////////////////////////////////////////////////////////
char *getSpecifiedKey(char code, KEY_TABLE *tbl, int numTable)
{
	static	char	keyname[128];
	char			*key;

	if( (key = matchKeyCode(code, tbl, numTable)) != NULL ) {
		strcpy(keyname, key);
	}else{
		sprintf(keyname, "(%d)", code);
	}

	return keyname;
}

//////////////////////////////////////////////////////////////
// Print usage
//////////////////////////////////////////////////////////////
// printUsage
//////////////////////////////////////////////////////////////
void printUsage()
{
	printf("Usage : hidconfig <OPTION> <VALUE1> <VALUE2> <VALUE3> <VALUE4> \n");
	printf("   <OPTION> -r Read current configuration. \n");
	printf("               The parameter of <VALUE1> to <VALUE4> are ignored.\n");
	printf("            -s Set configuration as <VALUE1> to <VALUE4>. \n");
	printf("\n");
	printf("   <VALUE?>    <VALUE?> is specified in the following format.\n");
	printf("                    KeyName,ModifiedKey \n");
	printf("\n");
	printf("               The KeyName must be selected from the following list.\n");
	printf("\n");
	printf("               KEY_A, KEY_B, KEY_C, KEY_D, KEF_E, KEY_F, KEY_G, KEY_H, KEY_I, \n");
	printf("               KEY_J, KEY_K, KEY_L, KEY_M, KEY_N, KEY_O, KEY_P, KEY_Q, KEY_R, \n");
	printf("               KEY_S, KEY_T, KEY_U, KEY_V, KEW_W, KEY_X, KEY_Y, KEY_Z, \n");
	printf("               KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, KEY_9,\n");
	printf("               KEY_F1, KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6, \n");
	printf("               KEY_F7, KEY_F8, KEY_F9, KEY_F10, KEY_F11, KEY_F12 \n");
	printf("\n");
	printf("               The ModifiedKey must be selected from the followin list. \n");
	printf("\n");
	printf("               CTRL_LEFT, SHIFT_LEFT, ALT_LEFT, CTRL_RIGHT, SHIFT_RIGHT, ALT_RIGHT\n");
	printf("\n");
	printf("               Note: <VALUE2> to <VALUE4> are for the 4Ch version.\n");
	printf("                     These are ignored in the 1Ch version.\n");
	printf("\n");
	printf("   Ex. \n");
	printf("      1) Read current configuration. \n");
	printf("            ./hidconfig -r \n");
	printf("      2) Set trigger event of F5 Key when trigger 1 is input.\n");
	printf("            ./hidconfig -s KEY_F5 \n");
    printf("      3) Set trigger event of R key + CTRL key when trigger 1 is input \n");
	printf("            ./hidconfig -s KEY_R,CTRL_LEFT \n");
	printf("      4) Assign the key from Trigger1 to Trigger4 \n");
	printf("            ./hidconfig -s KEY_L,SHIFT_LEFT KEY_R,SHIFT_LEFT, KEY_F5, KEY_F6 \n");
	printf("\n");
}


//////////////////////////////////////////////////////////////
// Parse command line argment
//////////////////////////////////////////////////////////////
int parseArg(int argc, char *argv[])
{
	int				mode = 0;
	int				i;
	int				maps = 0;
	char			*token;
	unsigned char	code;

	if( argc < 2 ) {
		return 0;
	}

	if( strcmp(argv[1], "-r") == 0 ) {
		mode = 1;
	}else if( strcmp(argv[1], "-s") == 0 ) {
		mode = 2;
		for(i = 2; i < argc; i++) {
			token = strtok(argv[i], ",");
			if(token != NULL ) {
				if( (code = getSpecifiedCode(token, tblKey, sizeof(tblKey)/sizeof(tblKey[0]))) != 0xFF ) {
					tblKeyMap[maps].keycode = code;
					if( (token = strtok(NULL, ",")) != NULL ) {
						if( (code = getSpecifiedCode(token, tblModifiedKey, sizeof(tblModifiedKey)/sizeof(tblModifiedKey[0]))) != 0xFF ) {
							tblKeyMap[maps].modifiedKey = code;
						}else{
							mode = 0;
							break;
						}
					}else{
						tblKeyMap[maps].modifiedKey = 0;
					}
					maps++;
					if( maps >= KEY_MAP_MAX )
						break;
				}else{
					mode = 0;
					break;
				}
			}
		}
		if( maps == 0 )
			mode = 0;
	}

	return mode;
}

//////////////////////////////////////////////////////////////
// Print current configuration
//////////////////////////////////////////////////////////////
// printConfiguration
//////////////////////////////////////////////////////////////
void printConfiguration(CONFIG_DATA *data)
{
	int		i;
	char	keyname[128];
	char	modifiedKey[128];

	printf("Current Configuration \n");
	printf("----------------------------------------------------\n");
	printf("  Ch    Key            Modified Key\n");
	printf("----------------------------------------------------\n");
	for(i = 0; i < KEY_MAP_MAX; i++) {
		if( data->key[i][1] == 0xFF )
			break;
		strcpy(keyname, getSpecifiedKey(data->key[i][1], tblKey, sizeof(tblKey)/sizeof(tblKey[0])));
		if( data->key[i][0] == 0 || data->key[i][0] == 0xFF ) {
			strcpy(modifiedKey, "");
		}else{
			strcpy(modifiedKey, getSpecifiedKey(data->key[i][0], tblModifiedKey, sizeof(tblModifiedKey)/sizeof(tblModifiedKey[0])));
		}
		printf(" Ch-%d   %-15s%s\n", i+1, keyname, modifiedKey);
	}
	printf("----------------------------------------------------\n");
	printf("\n");
}

//////////////////////////////////////////////////////////////
// Read current configuration from device
//////////////////////////////////////////////////////////////
// readConfiguration
//////////////////////////////////////////////////////////////
int	readConfiguration()
{
	usbDevice_t		*device = NULL;
	CONFIG_DATA		data;

	if( usbhidOpenDevice(&device, VENDOR_ID, NULL, PRODUCT_ID, NULL) != USBOPEN_SUCCESS ) {
		fprintf(stderr, "Device open error. \n");
		return 0;
	}

	data.reportId = 1;
	usbhidGetReport(device, (char *)&data, sizeof(CONFIG_DATA));

	usbhidCloseDevice(device);

	printConfiguration(&data);

	return 1;
}

//////////////////////////////////////////////////////////////
// Set configuration to device
//////////////////////////////////////////////////////////////
// SetConfiguration
//////////////////////////////////////////////////////////////
int	setConfiguration()
{
	usbDevice_t		*device = NULL;
	CONFIG_DATA		data;
	int				i;
	int				keys;

	memset(&data, 0xFF, sizeof(CONFIG_DATA));

	keys = 0;
	data.reportId = 1;
	for(i = 0; i < KEY_MAP_MAX; i++) {
		if( tblKeyMap[i].keycode == 0xFF )
			break;
		data.key[keys][1] = tblKeyMap[i].keycode;
		data.key[keys][0] = tblKeyMap[i].modifiedKey;
		keys++;
	}

	if( usbhidOpenDevice(&device, VENDOR_ID, NULL, PRODUCT_ID, NULL) != USBOPEN_SUCCESS ) {
		fprintf(stderr, "Device open error. \n");
		return 0;
	}

	usbhidSetReport(device, (char *)&data, sizeof(CONFIG_DATA));
	usbhidCloseDevice(device);

	return 1;
}

//////////////////////////////////////////////////////////////
// main
//////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
	int		mode;

	memset(tblKeyMap, 0xFF, sizeof(tblKeyMap));

	if((mode = parseArg(argc, argv)) == 0 ) {
		printUsage();
		exit(1);
	}

	switch( mode ) {
		case	1:			// Read config
			readConfiguration();
			break;
		case	2:			// Set config
			setConfiguration();
			break;
	}

	exit(0);
}


